#include "windows.h"
#include "extern.h"
#include <direct.h>
#include <time.h>
#include "sysintf.h"

typedef unsigned short word;
typedef unsigned long dword;

struct _stamp   /* DOS-style datestamp */
{
  struct
  {
    /* IBM Cset/2 is allergic to "unsigned short" when declaring bitfields! */

#ifdef __IBMC__
    unsigned int da : 5;
    unsigned int mo : 4;
    unsigned int yr : 7;
#else
/*lint -e46 */
    word da : 5;
    word mo : 4;
    word yr : 7;
/*lint -restore */
#endif
  } date;

  struct
  {
#ifdef __IBMC__
    unsigned int ss : 5;
    unsigned int mm : 6;
    unsigned int hh : 5;
#else
/*lint -e46 */
    word ss : 5;
    word mm : 6;
    word hh : 5;
/*lint -restore */
#endif
  } time;
};


struct _dos_st
{
  word date;
  word time;
};

/* Union so we can access stamp as "int" or by individual components */

union stamp_combo   
{
  dword ldate;
  struct _stamp msg_st;
  struct _dos_st dos_st;
};

typedef union stamp_combo SCOMBO;


/* Convert a DOS-style bitmapped date into a 'struct tm'-type date. */

static int is_dst=-1;

static void near InitCvt(void)
{
  time_t tnow;

  tnow=time(NULL);

  is_dst=!! (localtime(&tnow)->tm_isdst);
}

struct tm * DosDate_to_TmDate(union stamp_combo *dosdate,
                                     struct tm *tmdate)
{
  if (is_dst==-1)
    InitCvt();

  tmdate->tm_mday=dosdate->msg_st.date.da;
  tmdate->tm_mon =dosdate->msg_st.date.mo-1;
  tmdate->tm_year=dosdate->msg_st.date.yr+80;

  tmdate->tm_hour=dosdate->msg_st.time.hh;
  tmdate->tm_min =dosdate->msg_st.time.mm;
  tmdate->tm_sec =dosdate->msg_st.time.ss << 1;

  tmdate->tm_isdst=is_dst;

  return tmdate;
}




typedef struct ent {
   char    *name;
   time_t  mtime;
   int     isdir;
   struct ent *next;
} Entry, *EntryPtr;


typedef struct mydir {
   char         *path;
   uint32        hkey;
   EntryPtr      entries;
   struct mydir *next;
} DirEntry, *DirEntryPtr;

static DirEntryPtr dtab[HASH_TABLE_SIZE];


/* Stat a path using the directory cache.
 *
 * We build a cannonical representation of the path using either an absolute
 * path name if that is what 'path' is or the relative path name constructed
 * from 'path' and the present value of Pwd.
 *
 * The present value of Pwd then gives a directory path that we search for
 * in our cache using a hash lookup.  If the directory component is located
 * then we search the basename component of the path and return the result of
 * the search:  0L if the component is not in the cache and it's time stamp
 * otherwise.
 *
 * If the directory is not in our cache we insert it into the cache by
 * openning the directory and reading all of the files within.  Once read
 * then we return the result of the above search.
 *
 * Optionally, if force is TRUE, and we did NOT read the directory to provide
 * the result then stat the file anyway and update the internal cache.
 */

PUBLIC time_t
CacheStat(path, force)
char        *path;
int          force;
{
   struct stat stbuf;
   DirEntryPtr dp;
   EntryPtr    ep;
   uint32 hkey;
   uint16 hv;
   char *fpath;
   char *fpth;
   char *comp;
   char *dir;
   int  loaded=FALSE;

   if (If_root_path(path))
      fpth = path;
   else
      fpth = Build_path(Pwd,path);

   fpath = DmStrDup(fpth);
   strlwr(fpath);

   dir  = Filedir(DmStrDup(fpath));
   comp = DmStrDup(Basename(fpath));

   hv = Hash(dir,&hkey);

   for(dp=dtab[hv]; dp; dp=dp->next)
      if (hkey == dp->hkey && stricmp(dp->path,dir) == 0)
	 break;

   if (!dp) {
      DIR *dirp;
      struct dirent *direntp;

      if( Verbose & V_DIR_CACHE )
      {
         printf( "%s:  Caching directory [%s]\n", Pname, dir);
      }

      /* Load the directory, we have the right hash position already */
      loaded = TRUE;

      TALLOC(dp,1,DirEntry);
      dp->next = dtab[hv];
      dtab[hv] = dp;
      dp->path = DmStrDup(dir);
      dp->hkey = hkey;

      if (Set_dir(dir) == 0) {
         WIN32_FIND_DATA wfd;
         HANDLE hdir;
         char *temp = malloc(strlen(dir) + 5);

         if (!temp)
         {
          fprintf(stderr, "no mem for malloc temp in dcache\n");
          exit(1);
         }

         strcpy(temp, dir);

         if (*temp && temp[strlen(temp)-1] != '\\')
          strcat(temp, "\\");

         strcat(temp, "*.*");

         if ((hdir=FindFirstFile(temp, &wfd)) != INVALID_HANDLE_VALUE)
         {
           do
           {
             SCOMBO sc;
             struct tm tms;

             TALLOC(ep,1,Entry);
             ep->name = DmStrDup(wfd.cFileName);
             strlwr(ep->name);
             ep->next = dp->entries;
             dp->entries = ep;
             ep->isdir = !!(wfd.dwFileAttributes & 0x10);

              FileTimeToDosDateTime(&wfd.ftLastWriteTime,
                                    &sc.dos_st.date,
                                    &sc.dos_st.time);

             DosDate_to_TmDate(&sc, &tms);
             ep->mtime = mktime(&tms);
           }
           while (FindNextFile(hdir, &wfd));

           FindClose(hdir);
         }
         free(temp);
	 Set_dir(Pwd);
      }

      if( Verbose & V_DIR_CACHE )
      {
         printf( "%s:  Done caching directory [%s]\n", Pname, dir);
      }

   }

   if (dp) {
      for(ep=dp->entries; ep; ep=ep->next)
         if(stricmp(ep->name,comp) == 0)
	    break;
   }
   else
      ep = NULL;

   if( force && !loaded) {
      if( Verbose & V_DIR_CACHE )
	 printf("%s:  Updating dir cache entry for [%s]\n", Pname, fpath);

      if (strlen(comp) > NameMax || STAT(fpath,&stbuf) != 0) {
	 if(ep)
	    ep->mtime = 0L;
      }
      else {
	 if (!ep) {
	    TALLOC(ep,1,Entry);
	    ep->name = DmStrDup(comp);
            strlwr(ep->name);
	    ep->next = dp->entries;
	    ep->isdir = (stbuf.st_mode & S_IFDIR);
	    dp->entries = ep;
	 }

	 ep->mtime = stbuf.st_mtime;
      }
   }

   FREE(fpath);
   FREE(dir);
   FREE(comp);
   return((!ep || (Augmake && ep->isdir)) ? 0L : ep->mtime);
}
